const config = require("config");
const dayjs = require("dayjs");
const path = require("path");
const Key = require("./Key");
const RedisClient = require("./RedisClient");
const PaymentProcessor = require("./payment_processor/PaymentProcessor");

class Order {
  static get STORAGE_MUTEX_KEY_PREFIX() {
    return "order_mutex:";
  }

  static get STORAGE_KEY_PREFIX() {
    return "order:";
  }

  static get PURCHASE_TAX_AMOUNT() {
    return 0.07;
  }

  // How long is a link between order <=> key stored
  static get PURCHASE_LINK_TIME_DAYS() {
    return 31;
  }
  static get PURCHASE_LINK_ORDER_TO_KEY_PREFIX() {
    return "order_link_order_key:";
  }
  static get PURCHASE_LINK_KEY_TO_ORDERS_PREFIX() {
    return "order_link_key_order:";
  }

  // How many minutes is a user allowed to take for finishing the payment
  static get PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS() {
    return 6;
  }

  static GET_ORDER_FILE_BASE_PATH(order_date) {
    let order_path = config.get("storage.data_path");
    if (process.env.NODE_ENV === "development") {
      order_path = path.join(order_path, "development");
    }
    order_path = path.join(
      order_path,
      order_date.format("YYYY"),
      order_date.format("MM")
    );

    return order_path;
  }
  /**
   * Data stored in Redis Database
   */
  #order_id;
  #expires_at;
  #amount;
  #amount_refund_requested = 0;
  #amount_refunded = 0;
  #price;
  #order_key_charged = false;
  #payment_captured = false;

  #receipt_created = false;
  #civicrm_contribution_id;
  /** @type {PaymentProcessor} */
  #payment_processor;

  /**
   * Data populated by context
   */
  #order_date;
  #order_path;
  #redis_client;

  constructor(
    order_id,
    amount,
    price,
    payment_processor,
    amount_refund_requested = 0,
    order_key_charged = false,
    payment_captured = false,
    amount_refunded = 0,
    receipt_created = false,
    civicrm_contribution_id
  ) {
    this.#order_id = order_id;
    this.#expires_at = dayjs().add(6, "month");
    this.#order_date = dayjs.unix(this.#order_id.substr(0, 10));

    this.#order_path = Order.GET_ORDER_FILE_BASE_PATH(this.#order_date);

    this.#amount = parseInt(amount);
    this.#price = parseInt(price);

    this.#payment_processor = payment_processor;

    this.#amount_refund_requested = amount_refund_requested;
    this.#order_key_charged = order_key_charged;
    this.#payment_captured = payment_captured;
    this.#amount_refunded = amount_refunded;

    this.#receipt_created = receipt_created;

    this.#civicrm_contribution_id = civicrm_contribution_id;

    this.#redis_client = RedisClient.CLIENT();
  }

  getOrderID() {
    return this.#order_id;
  }

  getAmount() {
    return this.#amount;
  }

  getNettoPrice() {
    let amount = this.getPrice();
    let vat_amount = this.getVatAmount();
    return amount - vat_amount;
  }
  getVat() {
    return config.get("price.vat");
  }
  getVatAmount() {
    let vat = config.get("price.vat");
    let amount = this.getPrice();
    return amount * (vat / 100);
  }
  getPrice() {
    return this.#price;
  }

  /**
   *
   * @returns {PaymentProcessor}
   */
  getPaymentProcessor() {
    return this.#payment_processor;
  }

  async captureOrder() {
    if (this.isPaymentCaptured()) {
      throw new Error("Order already captured");
    }
    return this.getWriteLock()
      .then(() => this.getPaymentProcessor().captureOrder())
      .then(() => this.createCiviCRMOrder())
      .then(() => {
        this.#payment_captured = true;
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  async chargeKey() {
    if (this.isOrderKeyCharged()) {
      throw new Error("Key already charged");
    }
    return this.getWriteLock()
      .then(() => this.getKeyFromOrderLink())
      .then((key) => Key.CHARGE_KEY(key, this.getAmount()))
      .then(() => {
        this.#order_key_charged = true;
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  async requestRefund(amount) {
    if (this.#amount_refund_requested > 0) {
      throw new Error("Refund is already requested");
    }
    let new_charge = 0;
    return this.getWriteLock()
      .then(() => this.getKeyFromOrderLink())
      .then((key) => Key.DISCHARGE_KEY(key, amount))
      .then((charge) => {
        new_charge = charge;
        this.#amount_refund_requested = amount;
        return this.save();
      })
      .then(() => this.releaseWriteLock())
      .then(() => new_charge);
  }

  /**
   * Processes a refund which was previously issued by the customer.
   * Will issue a refund from the payment processor
   * 
   * @returns 
   */
  async refund() {
    if (this.#amount_refunded > 0 || this.#amount_refund_requested === 0) {
      throw "No refund requested or Refund already processed";
    }

    let refund_price = parseFloat(
      (
        Math.ceil(
          (this.#amount_refund_requested / this.#amount) * this.#price * 100
        ) / 100
      ).toFixed(2)
    );

    return this.getWriteLock()
      .then(() => this.getPaymentProcessor().refundOrder(refund_price))
      .then(() => {
        this.#amount_refunded = this.#amount_refund_requested;
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  /**
   * Processes an external refund. I.e. when the refund was issued through Payment Gateway directly
   * 
   * @param {float} price price that was refunded externally
   */
  async external_refund(price) {
    // Create amount from refunded price
    price = Math.min(this.#price, price);
    let amount = Math.floor((price / this.#price) * this.#amount);


    return this.getWriteLock().then(() => this.getKeyFromOrderLink()).then(key => Key.DISCHARGE_KEY(key, amount)).then(() => {
      this.#amount_refunded = amount;
      return this.save();
    }).then(() => this.releaseWriteLock());
  }

  async denyRefund() {
    if (this.#amount_refunded > 0 || this.#amount_refund_requested === 0) {
      return false;
    }
    return this.getWriteLock()
      .then(() => this.getKeyFromOrderLink())
      .then((key) => Key.CHARGE_KEY(key, this.#amount_refund_requested))
      .then(() => {
        this.#amount_refund_requested = 0;
        return this.save();
      })
      .then(() => this.releaseWriteLock())
      .then(() => {
        return true;
      })
      .catch((reason) => {
        console.debug(reason);
        return false;
      });
  }

  isReceiptCreated() {
    return this.#receipt_created;
  }

  getAmountRefundRequested() {
    return this.#amount_refund_requested;
  }

  isOrderKeyCharged() {
    return this.#order_key_charged;
  }
  isPaymentCaptured() {
    return this.#payment_captured;
  }
  getAmountRefunded() {
    return this.#amount_refunded;
  }

  /**
   *
   * @param {int} amount
   * @param {float} price
   * @param {string} key
   * @param {PaymentProcessor} payment_processor
   * @returns
   */
  static async CREATE_NEW_ORDER(amount, price, key, payment_processor) {
    return Order.GENERATE_UNIQUE_ORDER_ID().then((order_id) => {
      let new_order = new Order(order_id, amount, price, payment_processor);
      return new_order
        .getWriteLock()
        .then(() => new_order.getPaymentProcessor().createOrder(new_order)) // Create Order on PaymentProcessor side
        .then(() => new_order.createOrderLink(key))
        .then(() => new_order.save())
        .then(() => new_order.releaseWriteLock())
        .then(() => {
          return new_order;
        });
    });
  }

  static async LOAD_ORDER_FROM_ID(order_id) {
    let redis_client = RedisClient.CLIENT();
    return new Promise((resolve, reject) => {
      redis_client
        .hgetall(Order.STORAGE_KEY_PREFIX + order_id)
        .then((order_data) => {
          if (Object.keys(order_data).length === 0) {
            // Checking FS for order
            let order_date = dayjs.unix(order_id.substr(0, 10));
            let order_file = path.join(
              Order.GET_ORDER_FILE_BASE_PATH(order_date),
              order_id.toString() + ".json"
            );
            let fs = require("fs");
            if (fs.existsSync(order_file)) {
              order_data = JSON.parse(fs.readFileSync(order_file));
            } else {
              throw "Could not find Order in our database! Checking FS";
            }
          }
          let loaded_order = new Order(
            order_data.order_id,
            order_data.amount,
            order_data.price,
            PaymentProcessor.LOAD_PAYMENT_PROCESSOR(
              JSON.parse(order_data.payment_processor)
            ),
            JSON.parse(order_data.amount_refund_requested),
            JSON.parse(order_data.order_key_charged),
            JSON.parse(order_data.payment_captured),
            JSON.parse(order_data.amount_refunded),
            JSON.parse(order_data.receipt_created),
            order_data.civicrm_contribution_id
          );
          resolve(loaded_order);
        })
        .catch((reason) => {
          reject(reason);
        });
    });
  }

  async save() {
    let stored_data = {
      order_id: this.#order_id,
      expires_at: this.#expires_at.format("YYYY-MM-DD"),
      amount: this.#amount,
      price: this.#price,
      amount_refund_requested: this.#amount_refund_requested,
      order_key_charged: this.#order_key_charged,
      payment_captured: this.#payment_captured,
      amount_refunded: this.#amount_refunded,
      receipt_created: this.#receipt_created,
      payment_processor: JSON.stringify(this.#payment_processor.serialize()),
      civicrm_contribution_id: this.#civicrm_contribution_id,
    };
    /**
     * Completed Orders will be stored in Filesystem
     * Uncompleted Orders will be stored in Redis
     */
    let redis_key = Order.STORAGE_KEY_PREFIX + this.#order_id;
    if (this.isPaymentCaptured()) {
      let fs = require("fs");
      let order_file = path.join(
        this.#order_path,
        this.#order_id.toString() + ".json"
      );
      // Create directory if it does not exist
      if (!fs.existsSync(path.dirname(order_file))) {
        fs.mkdirSync(path.dirname(order_file), { recursive: true });
      }
      return this.#redis_client
        .del(redis_key)
        .then(() => this.updateOrderLinkTTL())
        .then(() =>
          fs.writeFileSync(order_file, JSON.stringify(stored_data, null, 4))
        );
    } else {
      // Store Order in Redis
      let expiration = new dayjs();
      expiration = expiration.add(
        Order.PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS,
        "hour"
      );
      return this.#redis_client
        .pipeline()
        .hmset(redis_key, stored_data)
        .expireat(redis_key, expiration.unix())
        .exec()
        .then(() => this.updateOrderLinkTTL());
    }
  }

  async delete() {
    let redis_key = Order.STORAGE_KEY_PREFIX + this.#order_id;
    if (!this.isPaymentCaptured()) {
      return this.deleteOrderLink()
        .then(() => this.#redis_client.del(redis_key))
        .then((deleted_keys) => {
          if (deleted_keys > 0) {
            return true;
          } else {
            return false;
          }
        })
        .catch(() => {
          return false;
        });
    } else {
      return false;
    }
  }

  async getWriteLock() {
    let write_lock_key = Order.STORAGE_MUTEX_KEY_PREFIX + this.#order_id;

    return new Promise(async (resolve, reject) => {
      let timestart = dayjs();
      let write_lock = 0;
      do {
        write_lock = await this.#redis_client
          .pipeline()
          .setnx(write_lock_key, 1)
          .expiretime(write_lock_key)
          .expire(write_lock_key, 15)
          .exec();
        let expire_at = write_lock[1][1];
        write_lock = write_lock[0][1];
        if (write_lock === 1) {
          resolve(true);
          break;
        } else if (dayjs().diff(timestart, "second") >= 15) {
          if (expire_at >= 0) {
            await this.#redis_client.expireat(write_lock_key, expire_at);
          }
          reject("Timed out waiting for write lock for Order");
          break;
        } else {
          if (expire_at >= 0) {
            await this.#redis_client.expireat(write_lock_key, expire_at);
          }
          await new Promise((resolve) => setTimeout(resolve, 1000));
        }
      } while (true);
    });
  }

  async releaseWriteLock() {
    let write_lock_key = Order.STORAGE_MUTEX_KEY_PREFIX + this.#order_id;
    return this.#redis_client.del(write_lock_key);
  }

  /**
   * Creates a link between Order and Key that will automatically expire
   * @param {string} key
   */
  async createOrderLink(key) {
    let expiration = dayjs().add(Order.PURCHASE_LINK_TIME_DAYS, "day");
    let redis_key_order_key =
      Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    let redis_key_key_order = Order.PURCHASE_LINK_KEY_TO_ORDERS_PREFIX + key;
    return this.#redis_client
      .pipeline()
      .set(redis_key_order_key, key)
      .expireat(redis_key_order_key, expiration.unix())
      .rpush(redis_key_key_order, this.#order_id)
      .expireat(redis_key_key_order, expiration.unix())
      .exec();
  }

  async deleteOrderLink() {
    let redis_key_order_key =
      Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    return this.getKeyFromOrderLink().then((key) => {
      let redis_key_key_order = Order.PURCHASE_LINK_KEY_TO_ORDERS_PREFIX + key;
      return this.#redis_client
        .pipeline()
        .del(redis_key_order_key)
        .lrem(redis_key_key_order, 0, this.#order_id)
        .exec();
    });
  }

  async createCiviCRMOrder() {
    let civicrm_data = config.get("app.civicrm");
    if (!civicrm_data.enabled) {
      return;
    }
    console.log(this.#order_date.format("YYYY-MM-DD HH:mm:ss"));
    return fetch(civicrm_data.url + "/Contribution/create", {
      method: "post",
      headers: {
        "Content-Type": "application/x-www-form-urlencoded",
        "X-Civi-Auth": `Bearer ${civicrm_data.api_key}`,
      },
      body:
        "params=" +
        encodeURIComponent(
          JSON.stringify({
            values: {
              contact_id: civicrm_data.user_id,
              financial_type_id: 5,
              receive_date: this.#order_date.format("YYYY-MM-DD HH:mm:ss"),
              total_amount: this.getPrice(),
              invoice_id: "INV_" + this.getOrderID(),
              invoice_number: "INV_" + this.getOrderID(),
              is_test: process.env.NODE_ENV === "development" ? true : false,
              is_pay_later: false,
              tax_amount: this.getVatAmount(),
              is_template: false,
            },
          })
        ),
    })
      .then((response) => {
        if (response.status !== 200) {
          throw "Error Creating Civicrm Contribution";
        } else {
          return response.json();
        }
      })
      .then((response_json) => {
        this.#civicrm_contribution_id = response_json.values[0].id;
      });
  }

  async attachReceipt(data) {
    let civicrm_data = config.get("app.civicrm");
    if (!civicrm_data.enabled) {
      return;
    }

    return this.getWriteLock()
      .then(() =>
        fetch(civicrm_data.url + "/Contribution/addReceipt", {
          method: "post",
          headers: {
            "Content-Type": "application/x-www-form-urlencoded",
            "X-Civi-Auth": `Bearer ${civicrm_data.api_key}`,
          },
          body:
            "params=" +
            encodeURIComponent(
              JSON.stringify({
                receipt: data,
                id: this.#civicrm_contribution_id,
              })
            ),
        })
      )
      .then((response) => {
        if (response.status !== 200) {
          throw "Error Creating Civicrm Contribution";
        } else {
          return response.json();
        }
      })
      .then((response_json) => {
        if (response_json["count"] !== 1) {
          throw JSON.stringify(response_json, null, 4);
        } else {
          this.#receipt_created = true;
          return response_json;
        }
      })
      .then(() => this.save())
      .then(() => this.releaseWriteLock());
  }

  async getReceipt() {
    let civicrm_data = config.get("app.civicrm");
    if (!civicrm_data.enabled) {
      return Promise.reject("No Civicrm Config supplied");
    }
    return fetch(civicrm_data.url + "/Contribution/getReceipt", {
      method: "post",
      headers: {
        "Content-Type": "application/x-www-form-urlencoded",
        "X-Civi-Auth": `Bearer ${civicrm_data.api_key}`,
      },
      body:
        "params=" +
        encodeURIComponent(
          JSON.stringify({
            id: this.#civicrm_contribution_id,
          })
        ),
    })
      .then((result) => result.json())
      .then((result_json) => {
        if (
          typeof result_json.error_code !== "undefined" &&
          result_json.error_code !== 0
        ) {
          throw result_json;
        } else {
          return result_json;
        }
      });
  }

  async updateOrderLinkTTL() {
    let redis_key = Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;

    let expiration = dayjs().add(Order.PURCHASE_LINK_TIME_DAYS, "day");
    if (!this.isPaymentCaptured()) {
      expiration = dayjs().add(
        Order.PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS,
        "hour"
      );
    }

    return this.#redis_client.expireat(redis_key, expiration.unix());
  }

  /**
   * Tries to get a key for this order. If the order link already expired this function rejects the promise
   */
  async getKeyFromOrderLink() {
    let redis_key = Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    return this.#redis_client.get(redis_key).then((result) => {
      if (result) {
        return result;
      } else {
        throw "Order Link does not exist";
      }
    });
  }

  static async GENERATE_UNIQUE_ORDER_ID() {
    // Generate Order ID => time in seconds since 1.1.1970 and and add a mutex to it to allow multiple order ids per second
    let Redis = require("ioredis");
    let redis_connection = new Redis({
      host: config.get("redis.host"),
    });
    let order_id = null;
    let order_base = Math.round(new Date().getTime() / 1000);
    let order_mutex = Math.floor(Math.random() * 10000);
    do {
      // make sure this order_id is not already registered
      order_id = "" + order_base + order_mutex;
      let order_lock = await redis_connection.setnx(order_id, true);
      if (order_lock === 1) {
        await redis_connection.expire(order_id, 5);
      } else {
        console.log("Couldn't acquire lock");
        order_mutex++;
      }
    } while (order_id === null);
    return order_id;
  }
}

module.exports = Order;
