const Crypto = require("../../app/Crypto.js");
const Order = require("../../app/Order.js");
const config = require("config");
const dayjs = require("dayjs");

var express = require("express");
var router = express.Router();
const { query, body, validationResult } = require("express-validator");

/* GET home page. */
router.get(
  "/",
  query("amount")
    .isInt({ min: -1, max: 12 })
    .withMessage("Amount needs to be between 0 and 4.")
    .toInt(),
  async function (req, res, next) {
    const errors = validationResult(req);
    if (!errors.isEmpty()) {
      return res.status(400).json({ errors: errors.array() });
    }

    /**
     * The user interface allows either 100 searches or steps of 250 searches (up to 12 * 250 = 3000)
     * 100 searches are a little bit more expensive than 250
     *
     * an amount of 0 corresponds to 100 searches
     * an amount of 1-12 corresponds to 1-12 * 250 searches
     */
    let params = {
      amount: req.query.amount === 0 ? 1 : req.query.amount,
      unit_size: req.query.amount === 0 ? 100 : 250,
      price_per_unit:
        req.query.amount === 0 ? Order.PRICE_FOR_100 : Order.PRICE_FOR_250,
      order_id: await generate_unique_order_id(),
      payments: {
        paypal: {
          client_id: config.get(`payments.paypal.client_id`),
        },
      },
    };

    let crypto = new Crypto();
    let order_date = dayjs();
    let expiration_date = order_date.add(
      Order.PURCHASE_STORAGE_TIME_MONTHS,
      "month"
    );
    let private_key = await crypto.private_key_get(order_date, expiration_date);
    params.crypto = {
      N: private_key.keyPair.n,
      E: private_key.keyPair.e,
    };
    params.expires_at = expiration_date.format("YYYY-MM-DD");

    // Generate hmac hash of the payment data so we are able to verify them when the client submits them again
    // Generate hmac hash of the payment data so we are able to verify them when the client submits them again
    params.integrity = crypto.createIntegrityHash(
      params.order_id,
      params.expires_at,
      params.amount,
      params.unit_size,
      params.price_per_unit,
      params.crypto.N,
      params.crypto.E
    );

    res.render("checkout/checkout", params);
  }
);

router.use(
  "/payment/order",
  body("amount")
    .isInt({ min: 1, max: 12 })
    .withMessage("Invalid amount submitted")
    .toInt(),
  body("unit_size").isIn(["100", "250"]).toInt(),
  body("price_per_unit")
    .isCurrency({ symbol: "", allow_negatives: false, thousands_separator: "" })
    .withMessage("Invalid Price Value.")
    .toFloat(),
  body("order_id").isInt({ min: 0 }).withMessage("Invalid `order_id` value"),
  body("expires_at")
    .isDate()
    .matches(/^\d{4}-\d{2}-\d{2}$/)
    .withMessage("Expiration date is not correct"),
  body("public_key_e")
    .isNumeric({ no_symbols: true })
    .withMessage("Invalid Public Key"),
  body("public_key_n")
    .isNumeric({ no_symbols: true })
    .withMessage("Invalid Public Key"),
  body("integrity")
    .isHash("sha256")
    .custom((value, { req }) => {
      let crypto = new Crypto();
      if (
        !crypto.validateIntegrityHash(
          value,
          req.body.order_id,
          req.body.expires_at,
          req.body.amount,
          req.body.unit_size,
          req.body.price_per_unit,
          req.body.public_key_n,
          req.body.public_key_e
        )
      ) {
        console.log("Invalid integrity");
        return Promise.reject("Integrity is not matching");
      }
      return true;
    })
    .withMessage(
      "Value for `integrity` must be a SHA256 Hash and validate against the purchase data."
    ),
  body("encrypted_sales_receipts")
    .custom((value, { req }) => {
      if (!Array.isArray(value)) {
        return Promise.reject("Parameter is not an array");
      }
      for (let i = 0; i < value.length; i++) {
        if (!/^\d+$/.test(value[i])) {
          return Promise.reject(
            "Encrypted Sales Receipt contains invalid data"
          );
        }
      }
      let expected_ticket_count = (req.body.amount * req.body.unit_size) / 50;
      if (expected_ticket_count % 1 !== 0) {
        return Promise.reject("Expected ticket count is not an integer.");
      }
      if (value.length !== expected_ticket_count) {
        return Promise.reject("Two many receipts compared to the order.");
      }
      return true;
    })
    .withMessage("Invalid Sales Receipt"),
  (req, res, next) => {
    // This route gets called when the user initiates a payment: Validate the submitted data here
    const errors = validationResult(req);
    if (!errors.isEmpty()) {
      return res.status(400).json({ errors: errors.array() });
    }

    next("route");
  }
);

/** Cancel is the same for all payment gateways */
router.post("/payment/order/*/cancel", (req, res) => {
  Order.LOAD_ORDER_FROM_ID(req.body.order_id).then((loaded_order) => {
    if (loaded_order.isPaymentComplete()) {
      res.status(400).json({
        msg: "Cannot delete a completed order",
      });
      return;
    }
    loaded_order.delete().then((success) => {
      if (success) {
        res.status(200).json({
          msg: "Order deleted",
        });
      } else {
        res.status(400).json({
          errors: [
            {
              msg: "Could not delete specified order",
            },
          ],
        });
      }
    });
  });
});

var developmentRouter = require("./development.js");
router.use("/payment/order/development", developmentRouter);

var paypalRouter = require("./paypal.js");
router.use("/payment/order/paypal", paypalRouter);

module.exports = router;
