var express = require("express");
var router = express.Router();
var multer = require('multer');
const { param, body, check, oneOf, validationResult, matchedData } = require("express-validator");
const config = require("config");


var orderRouter = require("./orders/orders");
var checkout_router_paypal = require("./checkout/paypal");

var Key = require("../app/Key");

router.get("/create", function (req, res, next) {
  Key.GET_NEW_KEY().then((key) => {
    res.redirect(`${res.locals.baseDir}/key/` + key + "?new=true");
  });
});

router.get("/remove", (req, res) => {
  if (req.cookies.key) {
    res.clearCookie("key");
  }
  res.redirect("/");
})

router.get("/enter", function (req, res, next) {
  if (req.cookies.key) {
    res.redirect(`${res.locals.baseDir}/key/` + encodeURIComponent(req.cookies.key));
  } else {
    res.render("login/key");
  }
});

const upload_storage = multer.memoryStorage();
const upload = multer({
  storage: upload_storage,
  limits: {
    fileSize: 5 * 1024 * 1024,
    files: 1,
    fields: 1
  }
});
router.post("/enter", upload.single('file'),
  body("key").optional({ checkFalsy: true }).exists().trim().isUUID(4).withMessage((value, { req }) => "Key is not valid"),
  function (req, res, next) {
    const errors = validationResult(req);
    if (!errors.isEmpty()) {
      res.render("login/key", { errors: errors.array() });
      return;
    }
    let req_data = matchedData(req);

    if (typeof req_data.key !== "undefined") {
      res.redirect(`${res.locals.baseDir}/key/` + req_data.key);
    } else if (typeof req.file === "undefined") {
      res.render("login/key", { errors: "File not provided or invalid" });
    } else {
      const jimp = require('jimp');

      jimp.read(req.file.buffer, (err, image) => {
        if (err) {
          res.render("login/key", { errors: ["Error reading image data"] });
          return;
        }
        const QrCode = require("qrcode-reader");
        let qr = new QrCode();
        qr.callback = (err, value) => {
          if (err) {
            res.render("login/key", { errors: ["Error decoding QR"] });
            return;
          }
          let url;
          try {
            url = new URL(value.result);
          } catch (err) {
            res.render("login/key", { errors: ["Error parsing URL"] });
            return;
          }
          let key = url.searchParams.get("key");
          if (key !== null) {
            res.redirect(`${res.locals.baseDir}/key/` + encodeURIComponent(key));
          } else {
            res.render("login/key", { errors: ["Error parsing URL"] });
          }
        };
        qr.decode(image.bitmap);
      });
    }

  });

router.use("/:key", param("key").isUUID(4), async (req, res, next) => {
  // Input Validation
  const errors = validationResult(req);
  if (!errors.isEmpty()) {
    return res.status(400).json({ errors: errors.array() });
  }

  let metager_url =
    "https://metager.de/meta/settings/load-settings?key=" +
    encodeURIComponent(req.params.key);
  let QRCode = require("qrcode");

  let qr_data_uri = await QRCode.toDataURL(metager_url, { errorCorrectionLevel: 'H', scale: 8 });

  req.data = Object.assign(req.data, {
    created_new: req.query.new === "true" ? true : false,
    key: {
      key: req.params.key,
      charge: await Key.GET_KEY_CHARGE(req.params.key),
      expiration: await Key.GET_EXPIRATION(req.params.key),
      settings_url: metager_url,
      qr: qr_data_uri,
    },
    cookies: req.cookies,
    page: "fill",
    links: {
      fill_url: `${res.locals.baseDir}/key/${req.params.key}`,
      orders_url: `${res.locals.baseDir}/key/${req.params.key}/orders`,
    },
    js: [],
    css: [`${res.locals.baseDir}/styles/key/key.css`],
  });

  next("route");
});

// Basic account page
router.get("/:key", async (req, res) => {
  if (req.data.admin) {
    res.redirect(`${res.locals.baseDir}/logout`);
    return;
  } else if (!req.cookies.key || req.cookies.key !== req.data.key.key) {
    res.cookie("key", req.data.key.key, {
      sameSite: "lax"
    });
  }
  res.render("key", req.data);
});

router.use("/:key/orders", orderRouter);

/**
 * Validate Amount field for checkout process
 */
router.use(
  "/:key/checkout/:amount?",
  param("amount")
    .optional({ checkFalsy: true })
    .isInt()
    .toInt()
    .isIn([300, 600, 900, 1200, 1800, 3600]),
  (req, res, next) => {
    // Input Validation
    const errors = validationResult(req);
    if (!errors.isEmpty()) {
      return res.status(400).json({ errors: errors.array() });
    }
    let amount = req.params.amount;
    if (typeof amount !== "undefined" && parseInt(req.params.amount) === 0) {
      amount = undefined;
    } else {
      amount = parseInt(amount);
    }
    req.data.checkout = {
      amount: amount,
    };

    // Add a URL to change the checkout amount
    req.data.change_url = {
      amount: `${res.locals.baseDir}/key/` + encodeURIComponent(req.data.key.key) + "#charge",
    };

    next("route");
  }
);

router.get("/:key/checkout/:amount", (req, res) => {
  if (req.query.error) {
    req.data.checkout.error = req.query.error;
  }
  req.data.js.push(`${res.locals.baseDir}/js/funding_sources.js`);
  res.render("key", req.data);
});

router.use("/:key/checkout/:amount?/paypal", checkout_router_paypal);

module.exports = router;
